import logging
from threading import Lock
from typing import Any, List, Dict, Set

from amplitude import Amplitude

from .config import LocalEvaluationConfig
from ..assignment import Assignment, AssignmentFilter, AssignmentService
from ..cohort.cohort import USER_GROUP_TYPE
from ..cohort.cohort_download_api import DirectCohortDownloadApi
from ..cohort.cohort_loader import CohortLoader
from ..cohort.cohort_storage import InMemoryCohortStorage
from ..deployment.deployment_runner import DeploymentRunner
from ..flag.flag_config_api import FlagConfigApiV2, FlagConfigStreamApi
from ..flag.flag_config_storage import InMemoryFlagConfigStorage
from ..user import User
from ..connection_pool import HTTPConnectionPool
from ..evaluation.engine import EvaluationEngine
from ..evaluation.topological_sort import topological_sort
from ..util import deprecated
from ..util.flag_config import get_grouped_cohort_ids_from_flags, get_all_cohort_ids_from_flag
from ..util.user import user_to_evaluation_context
from ..variant import Variant


class LocalEvaluationClient:
    """Experiment client for evaluating variants for a user locally."""

    def __init__(self, api_key: str, config: LocalEvaluationConfig = None):
        """
        Creates a new Experiment LocalEvaluationClient instance.
            Parameters:
                api_key (str): The environment API Key
                config (LocalEvaluationConfig): Config Object

            Returns:
                Experiment Client instance.
        """

        if not api_key:
            raise ValueError("Experiment API key is empty")
        self.engine = EvaluationEngine()
        self.api_key = api_key
        self.config = config or LocalEvaluationConfig()
        self.assignment_service = None
        if config and config.assignment_config:
            instance = Amplitude(config.assignment_config.api_key, config.assignment_config)
            self.assignment_service = AssignmentService(instance, AssignmentFilter(
                config.assignment_config.cache_capacity), config.assignment_config.send_evaluated_props)
        self.logger = logging.getLogger("Amplitude")
        self.logger.addHandler(logging.StreamHandler())
        if self.config.debug:
            self.logger.setLevel(logging.DEBUG)
        self.__setup_connection_pool()
        self.lock = Lock()
        self.cohort_storage = InMemoryCohortStorage()
        self.flag_config_storage = InMemoryFlagConfigStorage()
        cohort_loader = None
        if self.config.cohort_sync_config:
            cohort_download_api = DirectCohortDownloadApi(self.config.cohort_sync_config.api_key,
                                                          self.config.cohort_sync_config.secret_key,
                                                          self.config.cohort_sync_config.max_cohort_size,
                                                          self.config.cohort_sync_config.cohort_server_url,
                                                          self.logger)

            cohort_loader = CohortLoader(cohort_download_api, self.cohort_storage)
        flag_config_api = FlagConfigApiV2(api_key, self.config.server_url,
                                          self.config.flag_config_poller_request_timeout_millis)
        flag_config_stream_api = None
        if self.config.stream_updates:
            flag_config_stream_api = FlagConfigStreamApi(api_key, self.config.stream_server_url, self.config.stream_flag_conn_timeout)

        self.deployment_runner = DeploymentRunner(self.config, flag_config_api, flag_config_stream_api,
                                                  self.flag_config_storage, self.cohort_storage, self.logger,
                                                  cohort_loader)

    def start(self):
        """
        Fetch initial flag configurations and start polling for updates. You must call this function to begin
        polling for flag config updates.
        """
        self.deployment_runner.start()

    def evaluate_v2(self, user: User, flag_keys: Set[str] = None) -> Dict[str, Variant]:
        """
        Locally evaluates flag variants for a user.

        This function will only evaluate flags for the keys specified in the flag_keys argument. If flag_keys is
        missing or None, all flags are evaluated. This function differs from evaluate as it will return a default
        variant object if the flag was evaluated but the user was not assigned (i.e. off).

            Parameters:
                user (User): The user to evaluate
                flag_keys (List[str]): The flags to evaluate with the user. If empty, all flags are evaluated.

            Returns:
                The evaluated variants.
        """
        flag_configs = self.flag_config_storage.get_flag_configs()
        if flag_configs is None or len(flag_configs) == 0:
            return {}
        self.logger.debug(f"[Experiment] Evaluate: user={user} - Flags: {flag_configs}")
        sorted_flags = topological_sort(flag_configs, flag_keys and list(flag_keys))
        if not sorted_flags:
            return {}

        # Check if all required cohorts are in storage, if not log a warning
        self._required_cohorts_in_storage(sorted_flags)
        if self.config.cohort_sync_config:
            user = self._enrich_user_with_cohorts(user, flag_configs)

        context = user_to_evaluation_context(user)
        result = self.engine.evaluate(context, sorted_flags)
        variants = {
            k: Variant(
                key=v.key,
                value=v.value,
                payload=v.payload,
                metadata=v.metadata
            ) for k, v in result.items()
        }
        self.logger.debug(f"[Experiment] Evaluate Result: {variants}")
        if self.assignment_service is not None:
            self.assignment_service.track(Assignment(user, variants))
        return variants

    @deprecated("Use evaluate_v2")
    def evaluate(self, user: User, flag_keys: List[str] = None) -> Dict[str, Variant]:
        """
        Locally evaluates flag variants for a user.

        This function will only evaluate flags for the keys specified in the flag_keys argument. If flag_keys is
        missing, all flags are evaluated.

            Parameters:
                user (User): The user to evaluate
                flag_keys (List[str]): The flags to evaluate with the user. If empty, all flags are evaluated.

            Returns:
                The evaluated variants.
        """
        flag_keys = set(flag_keys) if flag_keys is not None else None
        variants = self.evaluate_v2(user, flag_keys)
        return self.__filter_default_variants(variants)

    def __setup_connection_pool(self):
        scheme, _, host = self.config.server_url.split('/', 3)
        timeout = self.config.flag_config_poller_request_timeout_millis / 1000
        self._connection_pool = HTTPConnectionPool(host, max_size=1, idle_timeout=30,
                                                   read_timeout=timeout, scheme=scheme)

    def stop(self) -> None:
        """
        Stop polling for flag configurations. Close resource like connection pool with client
        """
        self.deployment_runner.stop()
        self._connection_pool.close()

    def __enter__(self) -> 'LocalEvaluationClient':
        return self

    def __exit__(self, *exit_info: Any) -> None:
        self.stop()

    @staticmethod
    def __filter_default_variants(variants: Dict[str, Variant]) -> Dict[str, Variant]:
        def is_default_variant(variant: Variant) -> bool:
            default = False if variant.metadata.get('default') is None else variant.metadata.get('default')
            deployed = True if variant.metadata.get('deployed') is None else variant.metadata.get('deployed')
            return default or not deployed

        return {key: variant for key, variant in variants.items() if not is_default_variant(variant)}

    def _required_cohorts_in_storage(self, flag_configs: List) -> None:
        stored_cohort_ids = self.cohort_storage.get_cohort_ids()
        for flag in flag_configs:
            flag_cohort_ids = get_all_cohort_ids_from_flag(flag)
            missing_cohorts = flag_cohort_ids - stored_cohort_ids
            if missing_cohorts:
                message = (
                    f"Evaluating flag {flag.key} dependent on cohorts {flag_cohort_ids} "
                    f"without {missing_cohorts} in storage"
                    if self.config.cohort_sync_config
                    else f"Evaluating flag {flag.key} dependent on cohorts {flag_cohort_ids} without "
                         f"cohort syncing configured"
                )
                self.logger.warning(message)

    def _enrich_user_with_cohorts(self, user: User, flag_configs: Dict) -> User:
        grouped_cohort_ids = get_grouped_cohort_ids_from_flags(list(flag_configs.values()))

        if USER_GROUP_TYPE in grouped_cohort_ids:
            user_cohort_ids = grouped_cohort_ids[USER_GROUP_TYPE]
            if user_cohort_ids and user.user_id:
                user.cohort_ids = list(self.cohort_storage.get_cohorts_for_user(user.user_id, user_cohort_ids))

        if user.groups:
            for group_type, group_names in user.groups.items():
                group_name = group_names[0] if group_names else None
                if not group_name:
                    continue
                cohort_ids = grouped_cohort_ids.get(group_type, [])
                if not cohort_ids:
                    continue
                user.add_group_cohort_ids(
                    group_type,
                    group_name,
                    list(self.cohort_storage.get_cohorts_for_group(group_type, group_name, cohort_ids))
                )
        return user



import re
from dataclasses import dataclass
from typing import Optional

# Major and minor should be non-negative numbers separated by a dot
MAJOR_MINOR_REGEX = r'(\d+)\.(\d+)'

# Patch should be a non-negative number
PATCH_REGEX = r'(\d+)'

# Prerelease is optional. If provided, it should be a hyphen followed by a
# series of dot separated identifiers where an identifier can contain anything in [-0-9a-zA-Z]
PRERELEASE_REGEX = r'(-(([-\w]+\.?)*))?'

# Version pattern should be major.minor(.patchAndPreRelease) where .patchAndPreRelease is optional
VERSION_PATTERN = f'^{MAJOR_MINOR_REGEX}(\.{PATCH_REGEX}{PRERELEASE_REGEX})?$'

@dataclass
class SemanticVersion:
    major: int
    minor: int
    patch: int
    pre_release: Optional[str] = None

    @classmethod
    def parse(cls, version: Optional[str]) -> Optional['SemanticVersion']:
        """
        Parse a version string into a SemanticVersion object.
        Returns None if the version string is invalid.
        """
        if not version:
            return None

        match = re.match(VERSION_PATTERN, version)
        if not match:
            return None

        try:
            major = int(match.group(1))
            minor = int(match.group(2))
            patch = int(match.group(4)) if match.group(4) else 0
            pre_release = match.group(5) if match.group(5) else None
            return cls(major, minor, patch, pre_release)
        except (IndexError, ValueError):
            return None

    def compare_to(self, other: 'SemanticVersion') -> int:
        """
        Compare this version to another version.
        Returns:
            1 if this version is greater than the other
            -1 if this version is less than the other
            0 if the versions are equal
        """
        if self.major > other.major:
            return 1
        if self.major < other.major:
            return -1

        if self.minor > other.minor:
            return 1
        if self.minor < other.minor:
            return -1

        if self.patch > other.patch:
            return 1
        if self.patch < other.patch:
            return -1

        if self.pre_release and not other.pre_release:
            return -1
        if not self.pre_release and other.pre_release:
            return 1

        if self.pre_release and other.pre_release:
            if self.pre_release > other.pre_release:
                return 1
            if self.pre_release < other.pre_release:
                return -1
            return 0

        return 0

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SemanticVersion):
            return NotImplemented
        return self.compare_to(other) == 0



import threading
import time
import logging
from typing import Any

from http.client import HTTPConnection, HTTPResponse, HTTPSConnection


class WrapperHTTPConnection:

    def __init__(self, pool: 'HTTPConnectionPool', conn: HTTPConnection) -> None:
        """
        Wrapped Http Connection, used with connection pool
        :param pool: Connection pool this connection belongs to
        :param conn: Wrapped HTTPConnection
        """
        self.pool = pool
        self.conn = conn
        self.response = None
        self.last_time = time.time()
        self.is_available = True

    def __enter__(self) -> 'WrapperHTTPConnection':
        return self

    def __exit__(self, *exit_info: Any) -> None:
        if not self.response.will_close and not self.response.is_closed():
            self.close()
        self.pool.release(self)

    def request(self, *args: Any, **kwargs: Any) -> HTTPResponse:
        try:
            self.conn.request(*args, **kwargs)
            self.response = self.conn.getresponse()
            return self.response
        except Exception as e:
            self.close()
            raise e

    def close(self) -> None:
        self.conn.close()
        self.is_available = False


class HTTPConnectionPool:

    def __init__(self, host: str, port: int = None, max_size: int = None, idle_timeout: int = None,
                 read_timeout: float = None, scheme: str = 'https') -> None:
        """
        A simple connection pool to reuse the http connections
        :param host: pass
        :param port: pass
        :param max_size: Max connection allowed
        :param idle_timeout: Idle timeout to clear the connection
        :param read_timeout: Read timeout with connection
        :param scheme: http or https
        """
        self.host = host
        self.port = port
        self.max_size = max_size
        self.idle_timeout = idle_timeout
        self.read_timeout = read_timeout
        self.scheme = scheme
        self._lock = threading.Condition()
        self._pool = []
        self.conn_num = 0
        self.is_closed = False
        self._clearer = None
        self.start_clear_conn()

    def acquire(self, blocking: bool = True, timeout: int = None) -> WrapperHTTPConnection:
        if self.is_closed:
            raise ConnectionPoolClosed
        with self._lock:
            if self.max_size is None or not self.is_full():
                if self.is_pool_empty():
                    self._put_connection(self._create_connection())
            else:
                if not blocking:
                    if self.is_pool_empty():
                        raise EmptyPoolError
                elif timeout is None:
                    while self.is_pool_empty():
                        self._lock.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    end_time = time.time() + timeout
                    while self.is_pool_empty():
                        remaining = end_time - time.time()
                        if remaining <= 0:
                            raise EmptyPoolError
                        self._lock.wait(remaining)
            return self._get_connection()

    def release(self, conn: WrapperHTTPConnection) -> None:
        if self.is_closed:
            conn.close()
            return
        with self._lock:
            if not conn.is_available:
                conn.close()
                self.conn_num -= 1
                conn = self._create_connection()
            self._put_connection(conn)
            self._lock.notify()

    def _get_connection(self) -> WrapperHTTPConnection:
        try:
            return self._pool.pop()
        except IndexError:
            raise EmptyPoolError

    def _put_connection(self, conn: WrapperHTTPConnection) -> None:
        conn.last_time = time.time()
        self._pool.append(conn)

    def _create_connection(self) -> WrapperHTTPConnection:
        self.conn_num += 1
        connection = HTTPConnection if self.scheme == 'http:' else HTTPSConnection
        return WrapperHTTPConnection(self, connection(self.host, self.port, timeout=self.read_timeout))

    def is_pool_empty(self) -> bool:
        return len(self._pool) == 0

    def is_full(self) -> bool:
        if self.max_size is None:
            return False
        return self.conn_num >= self.max_size

    def close(self) -> None:
        if self.is_closed:
            return
        self.is_closed = True
        self.stop_clear_conn()
        pool, self._pool = self._pool, None
        for conn in pool:
            conn.close()

    def clear_idle_conn(self) -> None:
        if self.is_closed:
            raise ConnectionPoolClosed
        # Staring a thread to clear idle connections
        threading.Thread(target=self._clear_idle_conn).start()

    def _clear_idle_conn(self) -> None:
        if not self._lock.acquire(timeout=self.idle_timeout):
            return
        current_time = time.time()
        if self.is_pool_empty():
            pass
        elif current_time - self._pool[-1].last_time >= self.idle_timeout:
            self.conn_num -= len(self._pool)
            self._pool.clear()
        else:
            left, right = 0, len(self._pool) - 1
            while left < right:
                mid = (left + right) // 2
                if current_time - self._pool[mid].last_time >= self.idle_timeout:
                    left = mid + 1
                else:
                    right = mid
            self._pool = self._pool[left:]
            self.conn_num -= left
        self._lock.release()

    def start_clear_conn(self) -> None:
        if self.idle_timeout is None:
            return
        self.clear_idle_conn()
        self._clearer = threading.Timer(self.idle_timeout, self.start_clear_conn)
        self._clearer.daemon = True
        self._clearer.start()

    def stop_clear_conn(self) -> None:
        if self._clearer is not None:
            self._clearer.cancel()

    def __enter__(self) -> 'HTTPConnectionPool':
        return self

    def __exit__(self, *exit_info: Any) -> None:
        self.close()


class EmptyPoolError(Exception):
    pass


class ConnectionPoolClosed(Exception):
    pass



def hash_code(string) -> int:
    hash_value = 0
    if len(string) == 0:
        return hash_value
    for char in string:
        chr_code = ord(char)
        hash_value = ((hash_value << 5) - hash_value) + chr_code
        hash_value &= 0xFFFFFFFF
    return hash_value



from .remote.client import RemoteEvaluationClient
from .remote.config import RemoteEvaluationConfig
from .local.client import LocalEvaluationClient
from .local.config import LocalEvaluationConfig

remote_evaluation_instances = {}
local_evaluation_instances = {}


class Experiment:
    """Provides factory methods for storing instances of Client"""

    @staticmethod
    def initialize_remote(api_key: str, config: RemoteEvaluationConfig = None) -> RemoteEvaluationClient:
        """
        Initializes a remote evaluation client.
            Parameters:
                api_key (str): The Amplitude API Key
                config (RemoteEvaluationConfig): Optional Config

            Returns:
                A remote evaluation client.
        """
        if remote_evaluation_instances.get(api_key) is None:
            remote_evaluation_instances[api_key] = RemoteEvaluationClient(api_key, config)
        return remote_evaluation_instances[api_key]

    @staticmethod
    def initialize_local(api_key: str, config: LocalEvaluationConfig = None) -> LocalEvaluationClient:
        """
        Initialize a local evaluation client. A local evaluation client can evaluate local flags or experiments for a
        user without requiring a remote call to the amplitude evaluation server. In order to best leverage local
        evaluation, all flags, and experiments being evaluated server side should be configured as local.
            Parameters:
                api_key (str): The Amplitude API Key
                config (RemoteEvaluationConfig): Optional Config

            Returns:
                A local evaluation client.
        """
        if local_evaluation_instances.get(api_key) is None:
            local_evaluation_instances[api_key] = LocalEvaluationClient(api_key, config)
        return local_evaluation_instances[api_key]



import time


class Cache:
    class Node:
        def __init__(self, key, value):
            self.key = key
            self.value = value
            self.prev = None
            self.next = None
            self.last_accessed_time = time.time()

    def __init__(self, capacity, ttl_millis):
        self.capacity = capacity
        self.ttl_millis = ttl_millis
        self.cache = {}
        self.head = self.Node(None, None)
        self.tail = self.Node(None, None)
        self.head.next = self.tail
        self.tail.prev = self.head

    def _add_node(self, node):
        node.prev = self.head
        node.next = self.head.next
        self.head.next.prev = node
        self.head.next = node

    def _remove_node(self, node):
        prev = node.prev
        next_node = node.next
        prev.next = next_node
        next_node.prev = prev

    def _move_to_head(self, node):
        self._remove_node(node)
        self._add_node(node)

    def get(self, key):
        if key in self.cache:
            node = self.cache[key]
            current_time = time.time()
            if (current_time - node.last_accessed_time) * 1000 <= self.ttl_millis:
                node.last_accessed_time = current_time  # Update last accessed time
                self._move_to_head(node)
                return node.value
            else:
                # Node has expired, remove it from the cache
                self._remove_node(node)
                del self.cache[key]
        return None

    def put(self, key, value):
        if key in self.cache:
            node = self.cache[key]
            node.value = value
            node.last_accessed_time = time.time()  # Update last accessed time
            self._move_to_head(node)
        else:
            if len(self.cache) >= self.capacity:
                # Evict the least recently used node (tail's prev)
                tail_prev = self.tail.prev
                self._remove_node(tail_prev)
                del self.cache[tail_prev.key]

            new_node = self.Node(key, value)
            self._add_node(new_node)
            self.cache[key] = new_node



from dataclasses import dataclass
from typing import Dict, List, Optional, Any
from dataclasses_json import dataclass_json

from .select import selectable


@selectable
@dataclass_json
@dataclass
class EvaluationVariant:
    """Represents a variant in a feature flag evaluation."""
    key: Optional[str] = None
    value: Optional[Any] = None
    payload: Optional[Any] = None
    metadata: Optional[Dict[str, Any]] = None


@dataclass_json
@dataclass
class EvaluationDistribution:
    """Represents distribution configuration for a variant."""
    variant: str
    range: List[int]


@dataclass_json
@dataclass
class EvaluationAllocation:
    """Represents allocation configuration for bucketing."""
    range: List[int]
    distributions: List[EvaluationDistribution]


@dataclass_json
@dataclass
class EvaluationCondition:
    """Represents a condition for flag evaluation."""
    selector: List[str]
    op: str
    values: List[str]


@dataclass_json
@dataclass
class EvaluationBucket:
    """Represents bucketing configuration for a segment."""
    selector: List[str]
    salt: str
    allocations: List[EvaluationAllocation]


@dataclass_json
@dataclass
class EvaluationSegment:
    """Represents a segment in a feature flag."""
    bucket: Optional[EvaluationBucket] = None
    conditions: Optional[List[List[EvaluationCondition]]] = None
    variant: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None


@dataclass_json
@dataclass
class EvaluationFlag:
    """Represents a complete feature flag configuration."""
    key: str
    variants: Dict[str, EvaluationVariant]
    segments: List[EvaluationSegment]
    dependencies: Optional[List[str]] = None
    metadata: Optional[Dict[str, Any]] = None


class EvaluationOperator:
    """Constants for evaluation operators."""
    IS = 'is'
    IS_NOT = 'is not'
    CONTAINS = 'contains'
    DOES_NOT_CONTAIN = 'does not contain'
    LESS_THAN = 'less'
    LESS_THAN_EQUALS = 'less or equal'
    GREATER_THAN = 'greater'
    GREATER_THAN_EQUALS = 'greater or equal'
    VERSION_LESS_THAN = 'version less'
    VERSION_LESS_THAN_EQUALS = 'version less or equal'
    VERSION_GREATER_THAN = 'version greater'
    VERSION_GREATER_THAN_EQUALS = 'version greater or equal'
    SET_IS = 'set is'
    SET_IS_NOT = 'set is not'
    SET_CONTAINS = 'set contains'
    SET_DOES_NOT_CONTAIN = 'set does not contain'
    SET_CONTAINS_ANY = 'set contains any'
    SET_DOES_NOT_CONTAIN_ANY = 'set does not contain any'
    REGEX_MATCH = 'regex match'
    REGEX_DOES_NOT_MATCH = 'regex does not match'



from typing import Dict, Any


class Variant:
    """Variant Class"""

    def __init__(self, value: str = None, payload: Any = None, key: str = None, metadata: Dict[str, Any] = None):
        """
        Initialize a Variant
            Parameters:
                value (str): The value of the variant determined by the flag configuration.
                payload (Any): The attached payload, if any.
                key (str): The variant key.
                metadata (Dict[str, Any]: Additional variant metadata used by the system.

            Returns:
                An experiment variant
        """
        self.value = value
        self.payload = payload
        self.key = key
        self.metadata = metadata

    def __eq__(self, obj) -> bool:
        """
        Determine if current variant equal other variant
            Parameters:
                obj (Variant): The variant to compare with

            Returns:
                True if two variant equals, otherwise False
        """
        if obj is None:
            return False
        return self.key == obj.key and self.value == obj.value and self.payload == obj.payload

    def __str__(self):
        """Return Variant as string"""
        return f"key: {self.key}, value: {self.value}, payload: {self.payload}, metadata:{self.metadata}"



import json
import threading
from http.client import HTTPResponse, HTTPConnection, HTTPSConnection
from typing import List, Optional, Callable, Mapping, Union, Tuple

import sseclient

from ..connection_pool import HTTPConnectionPool
from ..util.updater import get_duration_with_jitter
from ..evaluation.types import EvaluationFlag
from ..version import __version__

class FlagConfigApi:
    def get_flag_configs(self) -> List[EvaluationFlag]:
        pass


class FlagConfigApiV2(FlagConfigApi):
    def __init__(self, deployment_key: str, server_url: str, flag_config_poller_request_timeout_millis: int):
        self.deployment_key = deployment_key
        self.server_url = server_url
        self.flag_config_poller_request_timeout_millis = flag_config_poller_request_timeout_millis
        self.__setup_connection_pool()

    def get_flag_configs(self) -> List[EvaluationFlag]:
        return self._get_flag_configs()

    def _get_flag_configs(self) -> List[EvaluationFlag]:
        conn = self._connection_pool.acquire()
        headers = {
            'Authorization': f"Api-Key {self.deployment_key}",
            'Content-Type': 'application/json;charset=utf-8',
            'X-Amp-Exp-Library': f"experiment-python-server/{__version__}"
        }
        body = None
        try:
            response = conn.request('GET', '/sdk/v2/flags?v=0', body, headers)
            response_body = response.read().decode("utf8")
            if response.status != 200:
                raise Exception(
                    f"[Experiment] Get flagConfigs - received error response: ${response.status}: ${response_body}")
            response_json = json.loads(response_body)
            return EvaluationFlag.schema().load(response_json, many=True)
        finally:
            self._connection_pool.release(conn)

    def __setup_connection_pool(self):
        scheme, _, host = self.server_url.split('/', 3)
        timeout = self.flag_config_poller_request_timeout_millis / 1000
        self._connection_pool = HTTPConnectionPool(host, max_size=1, idle_timeout=30,
                                                   read_timeout=timeout, scheme=scheme)


DEFAULT_STREAM_API_KEEP_ALIVE_TIMEOUT_MILLIS = 17000
DEFAULT_STREAM_MAX_CONN_DURATION_MILLIS = 15 * 60 * 1000
DEFAULT_STREAM_MAX_JITTER_MILLIS = 5000


class EventSource:
    def __init__(self, server_url: str, path: str, headers: Mapping[str, str], conn_timeout_millis: int,
                 max_conn_duration_millis: int = DEFAULT_STREAM_MAX_CONN_DURATION_MILLIS,
                 max_jitter_millis: int = DEFAULT_STREAM_MAX_JITTER_MILLIS,
                 keep_alive_timeout_millis: int = DEFAULT_STREAM_API_KEEP_ALIVE_TIMEOUT_MILLIS):
        self.keep_alive_timer: Optional[threading.Timer] = None
        self.server_url = server_url
        self.path = path
        self.headers = headers
        self.conn_timeout_millis = conn_timeout_millis
        self.max_conn_duration_millis = max_conn_duration_millis
        self.max_jitter_millis = max_jitter_millis
        self.keep_alive_timeout_millis = keep_alive_timeout_millis

        self.sse: Optional[sseclient.SSEClient] = None
        self.conn: Optional[HTTPConnection | HTTPSConnection] = None
        self.thread: Optional[threading.Thread] = None
        self._stopped = False
        self.lock = threading.RLock()

    def start(self, on_update: Callable[[str], None], on_error: Callable[[str], None]):
        with self.lock:
            if self.sse is not None:
                self.sse.close()
            if self.conn is not None:
                self.conn.close()

            self.conn, response = self._get_conn()
            if response.status != 200:
                on_error(f"[Experiment] Stream flagConfigs - received error response: ${response.status}: ${response.read().decode('utf-8')}")
                return

            self.sse = sseclient.SSEClient(response, char_enc='utf-8')
            self._stopped = False
            self.thread = threading.Thread(target=self._run, args=[on_update, on_error])
            self.thread.start()
            self.reset_keep_alive_timer(on_error)

    def stop(self):
        with self.lock:
            self._stopped = True
            if self.sse:
                self.sse.close()
            if self.conn:
                self.conn.close()
            if self.keep_alive_timer:
                self.keep_alive_timer.cancel()
            self.sse = None
            self.conn = None
            # No way to stop self.thread, on self.conn.close(),
            # the loop in thread will raise exception, which will terminate the thread.

    def reset_keep_alive_timer(self, on_error: Callable[[str], None]):
        with self.lock:
            if self.keep_alive_timer:
                self.keep_alive_timer.cancel()
            self.keep_alive_timer = threading.Timer(self.keep_alive_timeout_millis / 1000, self.keep_alive_timed_out,
                                                    args=[on_error])
            self.keep_alive_timer.start()

    def keep_alive_timed_out(self, on_error: Callable[[str], None]):
        with self.lock:
            if not self._stopped:
                self.stop()
                on_error("[Experiment] Stream flagConfigs - Keep alive timed out")

    def _run(self, on_update: Callable[[str], None], on_error: Callable[[str], None]):
        try:
            for event in self.sse.events():
                with self.lock:
                    if self._stopped:
                        return
                    self.reset_keep_alive_timer(on_error)
                if event.data == ' ':
                    continue
                on_update(event.data)
        except TimeoutError:
            # Due to connection max time reached, open another one.
            with self.lock:
                if self._stopped:
                    return
                self.stop()
                self.start(on_update, on_error)
        except Exception as e:
            # Closing connection can result in exception here as a way to stop generator.
            with self.lock:
                if self._stopped:
                    return
            on_error("[Experiment] Stream flagConfigs - Unexpected exception" + str(e))

    def _get_conn(self) -> Tuple[Union[HTTPConnection, HTTPSConnection], HTTPResponse]:
        scheme, _, host = self.server_url.split('/', 3)
        connection = HTTPConnection if scheme == 'http:' else HTTPSConnection

        body = None

        conn = connection(host, timeout=get_duration_with_jitter(self.max_conn_duration_millis, self.max_jitter_millis) / 1000)
        try:
            conn.request('GET', self.path, body, self.headers)
            response = conn.getresponse()
        except Exception as e:
            conn.close()
            raise e

        return conn, response


class FlagConfigStreamApi:
    def __init__(self,
                 deployment_key: str,
                 server_url: str,
                 conn_timeout_millis: int,
                 max_conn_duration_millis: int = DEFAULT_STREAM_MAX_CONN_DURATION_MILLIS,
                 max_jitter_millis: int = DEFAULT_STREAM_MAX_JITTER_MILLIS):
        self.deployment_key = deployment_key
        self.server_url = server_url
        self.conn_timeout_millis = conn_timeout_millis
        self.max_conn_duration_millis = max_conn_duration_millis
        self.max_jitter_millis = max_jitter_millis

        self.lock = threading.RLock()

        headers = {
            'Authorization': f"Api-Key {self.deployment_key}",
            'Content-Type': 'application/json;charset=utf-8',
            'X-Amp-Exp-Library': f"experiment-python-server/{__version__}"
        }

        self.eventsource = EventSource(self.server_url, "/sdk/stream/v1/flags", headers, conn_timeout_millis)

    def start(self, on_update: Callable[[List[EvaluationFlag]], None], on_error: Callable[[str], None]):
        with self.lock:
            init_finished_event = threading.Event()
            init_error_event = threading.Event()
            init_updated_event = threading.Event()

            def _on_update(data):
                response_json = json.loads(data)
                flags = EvaluationFlag.schema().load(response_json, many=True)
                if init_finished_event.is_set():
                    on_update(flags)
                else:
                    init_finished_event.set()
                    on_update(flags)
                    init_updated_event.set()

            def _on_error(data):
                if init_finished_event.is_set():
                    on_error(data)
                else:
                    init_error_event.set()
                    init_finished_event.set()
                    on_error(data)

            t = threading.Thread(target=self.eventsource.start, args=[_on_update, _on_error])
            t.start()
            init_finished_event.wait(self.conn_timeout_millis / 1000)
            if t.is_alive() or not init_finished_event.is_set() or init_error_event.is_set():
                self.stop()
                on_error("stream connection timeout error")
                return

            # Wait for first update callback to finish before returning.
            init_updated_event.wait()

    def stop(self):
        with self.lock:
            threading.Thread(target=lambda: self.eventsource.stop()).start()



import json
from typing import Dict, Any

from ..variant import Variant


def evaluation_variant_json_to_variant(variant_json: Dict[str, Any]) -> Variant:
    value = variant_json.get('value')
    if value is not None and type(value) != str:
        value = json.dumps(value, separators=(',', ':'))
    return Variant(
        value=value,
        payload=variant_json.get('payload'),
        key=variant_json.get('key'),
        metadata=variant_json.get('metadata')
    )


def evaluation_variants_json_to_variants(variants_json: Dict[str, Dict[str, Any]]) -> Dict[str, Variant]:
    variants: Dict[str, Variant] = {}
    for key, value in variants_json.items():
        variants[key] = evaluation_variant_json_to_variant(value)
    return variants



from typing import Dict, List, Optional

from .types import EvaluationFlag


class CycleException(Exception):
    """
    Raised when topological sorting encounters a cycle between flag dependencies.
    """

    def __init__(self, path: List[str]):
        self.path = path

    def __str__(self):
        return f"Detected a cycle between flags {self.path}"


def topological_sort(
        flags: Dict[str, EvaluationFlag],
        flag_keys: Optional[List[str]] = None
) -> List[EvaluationFlag]:
    """
    Perform a topological sort on feature flags based on their dependencies.
    """
    available: Dict[str, EvaluationFlag] = flags.copy()
    result: List[EvaluationFlag] = []
    starting_keys = flag_keys if flag_keys is not None else list(available.keys())

    for flag_key in starting_keys:
        traversal = _parent_traversal(flag_key, available)
        if traversal:
            result.extend(traversal)

    return result


def _parent_traversal(
        flag_key: str,
        available: Dict[str, EvaluationFlag],
        path: Optional[List[str]] = None
) -> Optional[List[EvaluationFlag]]:
    """
    Recursively traverse flag dependencies to build topologically sorted list.
    """
    if path is None:
        path = []

    flag = available.get(flag_key)
    if not flag:
        return None

    if not flag.dependencies or len(flag.dependencies) == 0:
        available.pop(flag.key)
        return [flag]

    path.append(flag.key)
    result: List[EvaluationFlag] = []

    for parent_key in flag.dependencies:
        if parent_key in path:
            raise CycleException(path)

        traversal = _parent_traversal(parent_key, available, path)
        if traversal:
            result.extend(traversal)

    result.append(flag)
    path.pop()
    available.pop(flag.key)
    return result



from .assignment import Assignment
from .assignment import DAY_MILLIS
from ..util.cache import Cache


class AssignmentFilter:
    def __init__(self, size: int, ttl_millis: int = DAY_MILLIS):
        self.cache = Cache(size, ttl_millis)

    def should_track(self, assignment: Assignment) -> bool:
        if not assignment.results:
            return False
        canonical_assignment = assignment.canonicalize()
        track = self.cache.get(canonical_assignment) is None
        if track:
            self.cache.put(canonical_assignment, object())
        return track



from amplitude import Amplitude, BaseEvent
from ..assignment.assignment import Assignment
from ..assignment.assignment import DAY_MILLIS
from ..assignment.assignment_filter import AssignmentFilter
from ..util import hash_code

FLAG_TYPE_MUTUAL_EXCLUSION_GROUP = "mutual-exclusion-group"
FLAG_TYPE_HOLDOUT_GROUP = "holdout-group"


def to_event(assignment: Assignment, send_evaluated_props: bool) -> BaseEvent:
    event = BaseEvent(event_type='[Experiment] Assignment', user_id=assignment.user.user_id,
                      device_id=assignment.user.device_id, event_properties={}, user_properties={})

    # If send_evaluated_props is True, populate event with all relevant user attributes
    if send_evaluated_props:
        user_attributes = [
            "country", "city", "region", "dma", "language",
            "platform", "version", "os", "device_manufacturer",
            "device_brand", "device_model", "carrier", "library", "ip_address"
        ]
        for attr in user_attributes:
            setattr(event, attr, getattr(assignment.user, attr, None))
        event.user_properties = assignment.user.user_properties

    set_props = {}
    unset_props = {}

    for flag_key in sorted(assignment.results):
        variant = assignment.results[flag_key]
        if variant.key is None:
            continue
        # Get variant metadata
        version: int = variant.metadata.get('flagVersion') if variant.metadata is not None else None
        segment_name: str = variant.metadata.get('segmentName') if variant.metadata is not None else None
        flag_type: str = variant.metadata.get('flagType') if variant.metadata is not None else None
        default: bool = False
        if variant.metadata is not None and variant.metadata.get('default') is not None:
            default = variant.metadata.get('default')
        # Set event properties
        event.event_properties[flag_key + '.variant'] = variant.key
        if version is not None and segment_name is not None:
            event.event_properties[flag_key + '.details'] = f"v{version} rule:{segment_name}"
        # Build user properties
        if flag_type == FLAG_TYPE_MUTUAL_EXCLUSION_GROUP:
            continue
        elif default:
            unset_props[f'[Experiment] {flag_key}'] = '-'
        else:
            set_props[f'[Experiment] {flag_key}'] = variant.key

    # Set user properties and insert id
    event.user_properties['$set'] = set_props
    event.user_properties['$unset'] = unset_props
    event.insert_id = f'{event.user_id} {event.device_id} {hash_code(assignment.canonicalize())} {int(assignment.timestamp / DAY_MILLIS)}'

    return event


class AssignmentService:
    def __init__(self, amplitude: Amplitude, assignment_filter: AssignmentFilter, send_evaluated_props: bool):
        self.amplitude = amplitude
        self.assignmentFilter = assignment_filter
        self.send_evaluated_props = send_evaluated_props

    def track(self, assignment: Assignment):
        if self.assignmentFilter.should_track(assignment):
            self.amplitude.track(to_event(assignment, self.send_evaluated_props))



import logging
from typing import Optional
import threading

from ..flag.flag_config_updater import FlagConfigPoller, FlagConfigStreamer, FlagConfigUpdaterFallbackRetryWrapper
from ..local.config import LocalEvaluationConfig
from ..cohort.cohort_loader import CohortLoader
from ..cohort.cohort_storage import CohortStorage
from ..flag.flag_config_api import FlagConfigApi, FlagConfigStreamApi
from ..flag.flag_config_storage import FlagConfigStorage
from ..local.poller import Poller
from ..util.flag_config import get_all_cohort_ids_from_flags

DEFAULT_STREAM_UPDATER_RETRY_DELAY_MILLIS = 15000
DEFAULT_STREAM_UPDATER_RETRY_DELAY_MAX_JITTER_MILLIS = 1000


class DeploymentRunner:
    def __init__(
            self,
            config: LocalEvaluationConfig,
            flag_config_api: FlagConfigApi,
            flag_config_stream_api: Optional[FlagConfigStreamApi],
            flag_config_storage: FlagConfigStorage,
            cohort_storage: CohortStorage,
            logger: logging.Logger,
            cohort_loader: Optional[CohortLoader] = None,
    ):
        self.config = config
        self.flag_config_api = flag_config_api
        self.flag_config_storage = flag_config_storage
        self.cohort_storage = cohort_storage
        self.cohort_loader = cohort_loader
        self.lock = threading.Lock()
        self.flag_updater = FlagConfigUpdaterFallbackRetryWrapper(
            FlagConfigPoller(flag_config_api, flag_config_storage, cohort_loader, cohort_storage, config, logger),
            None,
            0, 0, config.flag_config_polling_interval_millis, 0,
            logger
            )
        if flag_config_stream_api:
            self.flag_updater = FlagConfigUpdaterFallbackRetryWrapper(
                FlagConfigStreamer(flag_config_stream_api, flag_config_storage, cohort_loader, cohort_storage, logger),
                self.flag_updater,
                DEFAULT_STREAM_UPDATER_RETRY_DELAY_MILLIS, DEFAULT_STREAM_UPDATER_RETRY_DELAY_MAX_JITTER_MILLIS,
                config.flag_config_polling_interval_millis, 0,
                logger
            )

        self.cohort_poller = None
        if self.cohort_loader:
            self.cohort_poller = Poller(self.config.cohort_sync_config.cohort_polling_interval_millis / 1000,
                                        self.__update_cohorts)
        self.logger = logger

    def start(self):
        with self.lock:
            self.flag_updater.start(None)
            if self.cohort_loader:
                self.cohort_poller.start()

    def stop(self):
        self.flag_updater.stop()
        if self.cohort_poller:
            self.cohort_poller.stop()

    def __update_cohorts(self):
        cohort_ids = get_all_cohort_ids_from_flags(list(self.flag_config_storage.get_flag_configs().values()))
        try:
            self.cohort_loader.download_cohorts(cohort_ids).result()
        except Exception as e:
            self.logger.warning(f"Error while updating cohorts: {e}")



import json
import logging
import urllib.parse

from .user import User
import base64


class AmplitudeCookie:
    """This class provides utility functions for parsing and handling identity from Amplitude cookies."""

    @staticmethod
    def cookie_name(api_key: str, new_format: bool = False) -> str:
        """
        Get the cookie name that Amplitude sets for the provided
            Parameters:
                api_key (str): The Amplitude API Key
                new_format (bool): True if cookie is in Browser SDK 2.0 format

            Returns:
                The cookie name that Amplitude sets for the provided Amplitude API Key
       """
        if not api_key:
            raise ValueError("Invalid Amplitude API Key")

        if new_format:
            if len(api_key) < 10:
                raise ValueError("Invalid Amplitude API Key")
            return f"AMP_{api_key[0:10]}"

        if len(api_key) < 6:
            raise ValueError("Invalid Amplitude API Key")
        return f"amp_{api_key[0:6]}"

    @staticmethod
    def parse(amplitude_cookie: str, new_format: bool = False) -> User:
        """
        Parse a cookie string and returns user
            Parameters:
                amplitude_cookie (str):  A string from the amplitude cookie
                new_format: True if cookie is in Browser SDK 2.0 format

            Returns:
                Experiment User context containing a device_id and user_id (if available)
        """

        if new_format:
            decoding = base64.b64decode(amplitude_cookie).decode("utf-8")
            try:
                user_session = json.loads(urllib.parse.unquote_plus(decoding))
                if "userId" not in user_session:
                    return User(user_id=None, device_id=user_session["deviceId"])
                return User(user_id=user_session["userId"], device_id=user_session["deviceId"])
            except Exception as e:
                logger = logging.getLogger("Amplitude")
                logger.error("Error parsing the Amplitude cookie: " + str(e))
                return User()

        values = amplitude_cookie.split('.')
        user_id = None
        if values[1]:
            try:
                user_id = base64.b64decode(values[1]).decode("utf-8")
            except:
                user_id = None
        return User(user_id=user_id, device_id=values[0])

    @staticmethod
    def generate(device_id: str, new_format: bool = False) -> str:
        """
        Generates a cookie string to set for the Amplitude Javascript SDK
            Parameters:
                device_id (str):  A device id to set
                new_format: True if cookie is in Browser SDK 2.0 format

            Returns:
                A cookie string to set for the Amplitude Javascript SDK to read
        """
        if not new_format:
            return f"{device_id}.........."

        user_session_hash = {
            "deviceId": device_id
        }
        json_data = json.dumps(user_session_hash)
        encoded_json = urllib.parse.quote(json_data)
        base64_encoded = base64.b64encode(bytearray(encoded_json, "utf-8")).decode("utf-8")

        return base64_encoded



from dataclasses import dataclass, field
from typing import ClassVar, Set

USER_GROUP_TYPE: ClassVar[str] = "User"


@dataclass
class Cohort:
    id: str
    last_modified: int
    size: int
    member_ids: Set[str]
    group_type: str = field(default=USER_GROUP_TYPE)



"""
The official Amplitude Experiment Python Server SDK
.. include:: ../../README.md
"""

from .remote.client import RemoteEvaluationClient
from .remote.config import RemoteEvaluationConfig
from .variant import Variant
from .user import User
from .version import __version__
from .factory import Experiment
from .cookie import AmplitudeCookie
from .local.client import LocalEvaluationClient
from .local.config import LocalEvaluationConfig
from .local.config import ServerZone
from .assignment import AssignmentConfig
from .cohort.cohort_sync_config import CohortSyncConfig



import time
import logging
import base64
import json
from http.client import HTTPResponse
from typing import Optional
from ..version import __version__

from .cohort import Cohort
from ..connection_pool import HTTPConnectionPool
from ..exception import HTTPErrorResponseException, CohortTooLargeException

COHORT_REQUEST_RETRY_DELAY_MILLIS = 100


class CohortDownloadApi:

    def get_cohort(self, cohort_id: str, cohort: Optional[Cohort]) -> Cohort:
        raise NotImplementedError


class DirectCohortDownloadApi(CohortDownloadApi):
    def __init__(self, api_key: str, secret_key: str, max_cohort_size: int, server_url: str, logger: logging.Logger):
        super().__init__()
        self.api_key = api_key
        self.secret_key = secret_key
        self.max_cohort_size = max_cohort_size
        self.server_url = server_url
        self.logger = logger
        self.__setup_connection_pool()

    def get_cohort(self, cohort_id: str, cohort: Optional[Cohort]) -> Cohort or None:
        self.logger.debug(f"getCohortMembers({cohort_id}): start")
        errors = 0
        while True:
            response = None
            try:
                last_modified = None if cohort is None else cohort.last_modified
                response = self._get_cohort_members_request(cohort_id, last_modified)
                self.logger.debug(f"getCohortMembers({cohort_id}): status={response.status}")
                if response.status == 200:
                    cohort_info = json.loads(response.read().decode("utf8"))
                    self.logger.debug(f"getCohortMembers({cohort_id}): end - resultSize={cohort_info['size']}")
                    return Cohort(
                        id=cohort_info['cohortId'],
                        last_modified=cohort_info['lastModified'],
                        size=cohort_info['size'],
                        member_ids=set(cohort_info['memberIds']),
                        group_type=cohort_info['groupType'],
                    )
                elif response.status == 204:
                    self.logger.debug(f"getCohortMembers({cohort_id}): Cohort not modified")
                    return
                elif response.status == 413:
                    raise CohortTooLargeException(
                        f"Cohort exceeds max cohort size of {self.max_cohort_size}: {response.status}")
                elif response.status != 202:
                    raise HTTPErrorResponseException(response.status,
                                                     f"Unexpected response code: {response.status}")
            except Exception as e:
                if response and not (isinstance(e, HTTPErrorResponseException) and response.status == 429):
                    errors += 1
                self.logger.debug(f"getCohortMembers({cohort_id}): request-status error {errors} - {e}")
                if errors >= 3 or isinstance(e, CohortTooLargeException):
                    raise e
            time.sleep(COHORT_REQUEST_RETRY_DELAY_MILLIS / 1000)

    def _get_cohort_members_request(self, cohort_id: str, last_modified: int) -> HTTPResponse:
        headers = {
            'Authorization': f'Basic {self._get_basic_auth()}',
            'X-Amp-Exp-Library': f"experiment-python-server/{__version__}"
        }
        conn = self._connection_pool.acquire()
        try:
            url = f'/sdk/v1/cohort/{cohort_id}?maxCohortSize={self.max_cohort_size}'
            if last_modified is not None:
                url += f'&lastModified={last_modified}'
            response = conn.request('GET', url, headers=headers)
            return response
        finally:
            self._connection_pool.release(conn)

    def _get_basic_auth(self) -> str:
        credentials = f'{self.api_key}:{self.secret_key}'
        return base64.b64encode(credentials.encode('utf-8')).decode('utf-8')

    def __setup_connection_pool(self):
        scheme, _, host = self.server_url.split('/', 3)
        timeout = 10
        self._connection_pool = HTTPConnectionPool(host, max_size=10, idle_timeout=30, read_timeout=timeout,
                                                   scheme=scheme)



from enum import Enum

from ..assignment import AssignmentConfig
from ..cohort.cohort_sync_config import CohortSyncConfig, DEFAULT_COHORT_SYNC_URL, EU_COHORT_SYNC_URL

DEFAULT_SERVER_URL = 'https://api.lab.amplitude.com'
EU_SERVER_URL = 'https://flag.lab.eu.amplitude.com'

DEFAULT_STREAM_URL = 'https://stream.lab.amplitude.com'
EU_STREAM_SERVER_URL = 'https://stream.lab.eu.amplitude.com'


class ServerZone(Enum):
    US = "US"
    EU = "EU"


class LocalEvaluationConfig:
    """Experiment Local Client Configuration"""

    def __init__(self, debug: bool = False,
                 server_url: str = DEFAULT_SERVER_URL,
                 server_zone: ServerZone = ServerZone.US,
                 flag_config_polling_interval_millis: int = 30000,
                 flag_config_poller_request_timeout_millis: int = 10000,
                 stream_updates: bool = False,
                 stream_server_url: str = DEFAULT_STREAM_URL,
                 stream_flag_conn_timeout: int = 1500,
                 assignment_config: AssignmentConfig = None,
                 cohort_sync_config: CohortSyncConfig = None):
        """
        Initialize a config
           Parameters:
                debug (bool): Set to true to log some extra information to the console.
                server_url (str): The server endpoint from which to request flag configs.
                server_zone (ServerZone): Location of the Amplitude data center to get flags and cohorts from, US or EU
                flag_config_polling_interval_millis (int): The interval, in milliseconds, at which to poll for flag
                  configurations.
                flag_config_poller_request_timeout_millis (int): The request timeout, in milliseconds, used when
                  fetching flag configurations.
                assignment_config (AssignmentConfig): The assignment configuration.
                cohort_sync_config (CohortSyncConfig): The cohort sync configuration.

           Returns:
               The config object
        """
        self.debug = debug
        self.server_url = server_url
        self.server_zone = server_zone
        self.cohort_sync_config = cohort_sync_config
        if server_url == DEFAULT_SERVER_URL and server_zone == ServerZone.EU:
            self.server_url = EU_SERVER_URL
            if (cohort_sync_config is not None and
                    cohort_sync_config.cohort_server_url == DEFAULT_COHORT_SYNC_URL):
                self.cohort_sync_config.cohort_server_url = EU_COHORT_SYNC_URL

        self.stream_server_url = stream_server_url
        if stream_server_url == DEFAULT_SERVER_URL and server_zone == ServerZone.EU:
            self.stream_server_url = EU_STREAM_SERVER_URL

        self.flag_config_polling_interval_millis = flag_config_polling_interval_millis
        self.flag_config_poller_request_timeout_millis = flag_config_poller_request_timeout_millis
        self.stream_updates = stream_updates
        self.stream_flag_conn_timeout = stream_flag_conn_timeout
        self.assignment_config = assignment_config
